/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/. */

#include "gui.h"

static std::unique_ptr<Fl_Widget> selectedwidget;

using mcpppp::ui;
using mcpppp::paths;
using mcpppp::entries;
using mcpppp::deletedpaths;

// callback for run button
inline void run(Fl_Button* o, void* v)
{
	if (mcpppp::running)
	{
		return;
	}
	mcpppp::running = true;
	std::thread t(mcpppp::guirun);
	t.detach();
}

// callback for "CIM", "VMT", "FSB" checkboxes
inline void conversion(Fl_Check_Button* o, void* v)
{
	if (strcmp(o->label(), "FSB") == 0)
	{
		mcpppp::dofsb = static_cast<bool>(o->value());
	}
	if (strcmp(o->label(), "VMT") == 0)
	{
		mcpppp::dovmt = static_cast<bool>(o->value());
	}
	if (strcmp(o->label(), "CIM") == 0)
	{
		mcpppp::docim = static_cast<bool>(o->value());
	}
}

// callback for resourcepack checkboxes
inline void resourcepack(Fl_Check_Button* o, void* v)
{
	entries.at(static_cast<size_t>(*(static_cast<int*>(v)))).first = static_cast<bool>(o->value());
	ui->allpacks->value(0);
}

// callback for browse button
inline void browse(Fl_Button* o, void* v)
{
	ui->edit_paths->show();
}

// callback for reload button
inline void reload(Fl_Button* o, void* v)
{
	entries.clear();
	ui->scroll->clear();
	mcpppp::numbuttons = 0;
	ui->scroll->begin(); // padding
	std::unique_ptr<Fl_Check_Button> pad = std::make_unique<Fl_Check_Button>(470, 45, 150, 15);
	pad->down_box(FL_DOWN_BOX);
	pad->labeltype(FL_NO_LABEL);
	pad->hide();
	pad->deactivate();
	ui->scroll->end();
	for (const std::string& path : paths)
	{
		if (std::filesystem::is_directory(std::filesystem::u8path(path)))
		{
			for (const auto& entry : std::filesystem::directory_iterator(std::filesystem::u8path(path)))
			{
				if (entry.is_directory() || entry.path().extension() == ".zip")
				{
					entries.push_back(std::make_pair(true, entry));
					mcpppp::addpack(entry.path().filename().u8string(), true);
				}
			}
		}
	}
	ui->allpacks->value(1);
	Fl::wait();
}

// callback for path_input
inline void editpath(Fl_Input* o, void* v)
{
	deletedpaths.insert(paths.begin(), paths.end());
	paths.clear();
	std::string str = o->value();
	while (str.find(" // ") != std::string::npos)
	{
		const size_t i = str.find(" // ");
		paths.insert(std::string(str.begin(), str.begin() + static_cast<std::string::difference_type>(i)));
		deletedpaths.erase(std::string(str.begin(), str.begin() + static_cast<std::string::difference_type>(i)));
		str.erase(str.begin(), str.begin() + static_cast<std::string::difference_type>(i + 4));
	}
	if (!str.empty())
	{
		paths.insert(str);
		deletedpaths.erase(str);
	}
	mcpppp::addpaths();
	mcpppp::updatepathconfig();
}

// callback for "Add" button in "Edit Paths"
inline void addrespath(Fl_Button* o, void* v)
{
	std::string str;
#ifdef _WIN32
	str = mcpppp::winfilebrowser();
#else
	std::unique_ptr<Fl_Native_File_Chooser> chooser = std::make_unique<Fl_Native_File_Chooser>(1); // browse directory
	str = chooser->show();
#endif
	if (!str.empty() && paths.find(str) == paths.end())
	{
		mcpppp::addpath(str);
		deletedpaths.erase(str);
		mcpppp::updatepaths();
		mcpppp::updatepathconfig();
	}
	ui->edit_paths->redraw();
	reload(nullptr, nullptr);
}

// callback for "Delete" button in "Edit Paths"
inline void deleterespath(Fl_Button* o, void* v)
{
	if (selectedwidget == nullptr)
	{
		return;
	}
	std::string s = selectedwidget->label();
	// erase spaces used for padding
	s.erase(s.begin(), s.begin() + 4);
	paths.erase(s);
	deletedpaths.insert(s);
	selectedwidget.reset();
	mcpppp::addpaths();
	mcpppp::updatepaths();
	ui->paths->hide();
	ui->paths->show();
	mcpppp::updatepathconfig();
	ui->edit_paths->redraw();
	reload(nullptr, nullptr);
}

// callback for paths buttons in "Edit Paths"
inline void selectpath(Fl_Radio_Button* o, void* v) noexcept
{
	selectedwidget.reset(o);
}

// callback for settings button
inline void opensettings(Fl_Button* o, void* v)
{
	ui->settings->show();
	ui->box1->redraw(); // the outlining boxes disappear for some reason
	ui->box2->redraw();
}

// callback for help button
inline void openhelp(Fl_Button* o, void* v)
{
	ui->help->show();
	ui->box1->redraw();
	ui->box2->redraw();
}

// callback for save button in setings
inline void savesettings(Fl_Button* o, void* v)
{
	using mcpppp::config;

	// TODO: automate addition of settings here?
	// find some way to store pointer to widgets/value() functions in gui but not cli
	config["gui"]["settings"]["autoDeleteTemp"] = static_cast<bool>(ui->autodeletetemptrue->value());
	config["gui"]["settings"]["log"] = ui->log->value();
	config["gui"]["settings"]["timestamp"] = static_cast<bool>(ui->timestamptrue->value());
	config["gui"]["settings"]["outputLevel"] = ui->outputlevel->value();
	config["gui"]["settings"]["logLevel"] = ui->loglevel->value();
	config["gui"]["settings"]["autoReconvert"] = static_cast<bool>(ui->autoreconverttrue->value());
	config["gui"]["settings"]["fsbTransparent"] = static_cast<bool>(ui->fsbtransparenttrue->value());

	// remove excess settings
	std::vector<std::string> toremove;
	std::ifstream configfile("mcpppp-config.json");
	std::string temp;
	nlohmann::json j;
	try
	{
		temp.resize(std::filesystem::file_size("mcpppp-config.json"));
		configfile.read(&temp.at(0), static_cast<std::streamsize>(std::filesystem::file_size("mcpppp-config.json")));
		j = nlohmann::json::parse(temp, nullptr, true, true);
	}
	catch (const nlohmann::json::exception& e)
	{
		out(5) << e.what() << std::endl;
		throw e;
	}
	configfile.close();
	// remove duplicate settings (settings that are the same in gui and non-gui config)
	if (j.contains("settings"))
	{
		if (j["settings"].type() == nlohmann::json::value_t::object)
		{
			for (const auto& setting : j["settings"].items())
			{
				for (const auto& setting2 : config["gui"]["settings"].items())
				{
					if (mcpppp::lowercase(setting.key()) == mcpppp::lowercase(setting2.key()))
					{
						if (setting.value() == nlohmann::json(setting2.value()))
						{
							config["gui"]["settings"].erase(setting2.key());
							break;
						}
					}
				}
			}
		}
	}
	// remove default settings
	for (const auto& setting : config["gui"]["settings"].items())
	{
		if (mcpppp::settings.at(mcpppp::lowercase(setting.key())).default_val == nlohmann::json(setting.value()))
		{
			toremove.push_back(setting.key());
		}
	}
	for (const std::string& s : toremove)
	{
		config["gui"]["settings"].erase(s);
	}

	std::ofstream fout("mcpppp-config.json");
	fout << "// Please check out the Documentation for the config file before editing it yourself: https://github.com/supsm/MCPPPP/blob/master/CONFIG.md" << std::endl;
	fout << config.dump(1, '\t') << std::endl;
	fout.close();

	mcpppp::readconfig();

	ui->savewarning->hide();
}

// callback for edited settings
inline void settingchanged(Fl_Widget* o, void* v)
{
	ui->savewarning->show();
}

// callback for select all/none
inline void selectall(Fl_Check_Button* o, void* v)
{
	ui->scroll->clear();
	mcpppp::numbuttons = 0;
	// padding
	std::unique_ptr<Fl_Check_Button> pad = std::make_unique<Fl_Check_Button>(470, 45, 150, 15);
	pad->down_box(FL_DOWN_BOX);
	pad->labeltype(FL_NO_LABEL);
	pad->hide();
	pad->deactivate();
	ui->scroll->add(pad.get());
	pad.release();
	for (auto& entry : entries)
	{
		mcpppp::addpack(entry.second.path().filename().u8string(), static_cast<bool>(o->value()));
		entry.first = static_cast<bool>(o->value());
	}
	ui->scroll->redraw();
	Fl::wait();
}

// callback for closing main window
inline void windowclosed(Fl_Double_Window* w, void* v)
{
	if (mcpppp::running)
	{
		switch (fl_choice("Conversion is still in progress, files may be corrupted if you choose to close now. Are you sure you want to close?", "Yes", "No", nullptr))
		{
		case 0: // if yes, close window
			w->hide();
			break;
		case 1: // don't do anything
			break;
		default:
			break;
		}
	}
	else
	{
		w->hide();
	}
}
